@@ -45,7 +45,7 @@
 %%%
 %%% @author Michael Truog <mjtruog [at] gmail (dot) com>
 %%% @copyright 2011-2016 Michael Truog
-%%% @version 1.5.2 {@date} {@time}
+%%% @version 1.5.4 {@date} {@time}
 %%%------------------------------------------------------------------------
 
 -module(cloudi_service_api).
@@ -69,6 +69,7 @@
          nodes_alive/1,
          nodes_dead/1,
          nodes/1,
+         logging_set/2,
          logging_file_set/2,
          logging_level_set/2,
          logging_syslog_set/2,
@@ -309,12 +310,21 @@
 -type owner_external() ::
     list({user, pos_integer() | string()} |
          {group, pos_integer() | string()}).
+-type nice_external() ::
+    -20..20.
+-type cgroup_external() ::
+    undefined |
+    list({name, nonempty_string()} |
+         {parameters, list({nonempty_string(), string()})} |
+         {update_or_create, boolean()}).
 -type directory_external() ::
     file:filename() | undefined.
 -export_type([limit_external_key/0,
               limit_external_value/0,
               limit_external/0,
               owner_external/0,
+              nice_external/0,
+              cgroup_external/0,
               directory_external/0]).
 
 -type service_options_internal() ::
@@ -1148,6 +1158,23 @@ nodes(Timeout)
           (Timeout =:= infinity)) ->
     cloudi_core_i_nodes:nodes(Timeout).
 
+%%-------------------------------------------------------------------------
+%% @doc
+%% ===Set CloudI logging configuration.===
+%% @end
+%%-------------------------------------------------------------------------
+
+-spec logging_set(L :: logging_proplist(),
+                  Timeout :: api_timeout_milliseconds()) ->
+    ok | {error, file:posix() | badarg | system_limit}.
+
+logging_set([_ | _] = L, Timeout)
+    when ((is_integer(Timeout) andalso
+           (Timeout > ?TIMEOUT_DELTA) andalso
+           (Timeout =< ?TIMEOUT_MAX_ERLANG)) orelse
+          (Timeout =:= infinity)) ->
+    cloudi_core_i_configurator:logging_set(L, Timeout).
+
 %%-------------------------------------------------------------------------
 %% @doc
 %% ===Modify the current log file path.===
